--
--
local _floor  = math.floor
local _max    = math.max
local _random = math.random
local _sin    = math.sin
local _cos    = math.cos
local _sqrt   = math.sqrt
local _insert = table.insert
local _remove = table.remove

local _pi = math.pi

local maxResults           = 5000
local maxAttemptsPerSample = 30
local minRadius            = 10
local maxRadius            = 0
local stepRadius           = 0
local spawnRadiusFactor    = 2
local sizeW                = 800
local sizeH                = 600
local samples              = {}
local spawn                = {}
local spawnType            = 3

local hashSize  = 0
local hashTable = {}

local radii = {}

-- Apply settings table to current poisson instance
local function ApplySettings(s)
    maxResults           = s.maxResults or maxResults
    maxAttemptsPerSample = s.maxAttemptsPerSample or maxAttemptsPerSample
    minRadius            = s.minRadius or minRadius
    maxRadius            = s.maxRadius or maxRadius
    stepRadius           = s.stepRadius or stepRadius
    spawnRadiusFactor    = s.spawnRadiusFactor or spawnRadiusFactor
    sizeW                = s.sizeW or sizeW
    sizeH                = s.sizeH or sizeH
    samples              = s.samples or samples
    spawn                = s.spawn or spawn
    spawnType            = s.spawnType or spawnType
end

-- Get setting table from current poisson instance
local function GetSettings()
    return {
        maxResults           = maxResults,
        maxAttemptsPerSample = maxAttemptsPerSample,
        minRadius            = minRadius,
        maxRadius            = maxRadius,
        stepRadius           = stepRadius,
        spawnRadiusFactor    = spawnRadiusFactor,
        sizeW                = sizeW,
        sizeH                = sizeH,
        samples              = samples,
        spawn                = spawn,
        spawnType            = spawnType,
    }
end

local function _sanatizeSettings()
    maxResults           = _max(1, maxResults)
    minRadius            = _max(1, minRadius)
    maxRadius            = _max(minRadius, maxRadius)
    maxAttemptsPerSample = _max(1, maxAttemptsPerSample)
    sizeW                = sizeW > 0 and sizeW or maxRadius
    sizeH                = sizeH > 0 and sizeH or maxRadius
    spawnRadiusFactor    = spawnRadiusFactor > 1 and spawnRadiusFactor or 1.01
    stepRadius           = maxRadius - minRadius >= stepRadius
        and stepRadius
        or (maxRadius - minRadius)
    samples              = {}
    spawn                = {}
    hashTable            = {}
end

local function _computeRadii()
    radii = {}
    if stepRadius == 0 then
        _insert(radii, minRadius)
    else
        for i = minRadius, maxRadius, stepRadius do
            _insert(radii, i)
        end
    end
    if radii[#radii] ~= maxRadius then
        _insert(radii, maxRadius)
    end
end

local function _dist2(p1, p2)
    return (p1.x - p2.x) * (p1.x - p2.x) + (p1.y - p2.y) * (p1.y - p2.y)
end

local function _getHash(x, y)
    return "x" .. _floor(x / hashSize) .. "y" .. _floor(y / hashSize)
end

local function _getSurroundingBuckets(x, y)
    local cells = {}
    --tleft
    _insert(cells, _getHash(x - hashSize, y - hashSize))
    --tcenter
    _insert(cells, _getHash(x, y - hashSize))
    --tright
    _insert(cells, _getHash(x + hashSize, y - hashSize))
    --cleft
    _insert(cells, _getHash(x - hashSize, y))
    --c
    _insert(cells, _getHash(x, y))
    --cright
    _insert(cells, _getHash(x + hashSize, y))
    --bleft
    _insert(cells, _getHash(x - hashSize, y + hashSize))
    --bcenter
    _insert(cells, _getHash(x, y + hashSize))
    --bright
    _insert(cells, _getHash(x + hashSize, y + hashSize))
    return cells
end

local spawnMethods = require("poisson_spawn")
local function _spawnMethod(n)
    spawnMethods(spawn, _random)
    if type(n) == "string" then
        for _, value in pairs(spawnMethods) do
            if value.type == n then
                return value.exec()
            end
        end
    elseif type(n) == "number" then
        return spawnMethods[n].exec()
    else
        return 1
    end
end

local function _isValid(point)
    -- TODO Allow other padding types
    -- Currently always padded (no point can be closer then its radius to an edge)
    if (point.x >= 0 + point.r and
        point.x <= sizeW - point.r and
        point.y >= 0 + point.r and
        point.y <= sizeH - point.r)
    then
        local hashList = _getSurroundingBuckets(point.x, point.y)
        for _, hash in pairs(hashList) do
            if hashTable[hash] then
                for _, otherPoint in pairs(hashTable[hash]) do
                    local r = _max(point.r, otherPoint.r)
                    if _dist2(point, otherPoint) < r * r then
                        return false
                    end
                end
            end
        end
        return true
    end
end

local function Generate()
    _sanatizeSettings()
    _computeRadii()

    hashSize = maxRadius / _sqrt(2) * 2

    -- need some initial points to start from
    if #spawn == 0 then
        for i = 1, _random(2, 10) do
            _insert(spawn, {
                x = sizeW * _random(),
                y = sizeH * _random(),
                r = radii[_random(1, #radii)]
            })
        end
    end

    while (#spawn > 0) and (#samples < maxResults) do

        -- _spawnMethod (the function and related module) can be removed if you don't
        -- plan on ascribing meta meaning (say age, or color) based on the order of sample
        -- points - the methed below can be replaced with _random(1,#spawn)
        local spawnIdx   = _spawnMethod(spawnType) --_random(1,#spawn)
        local spawnPoint = spawn[spawnIdx]

        local pointFound = false

        for i = 1, maxAttemptsPerSample do
            local angle  = _random() * _pi * 2
            local dir    = { x = _cos(angle), y = _sin(angle) }
            local radius = radii[_random(1, #radii)]

            local min       = _max(spawnPoint.r, radius)
            local max       = min * spawnRadiusFactor
            local candidate = {
                x = (spawnPoint.x + dir.x * _random(min, max)),
                y = (spawnPoint.y + dir.y * _random(min, max)),
                r = radius
            }

            if _isValid(candidate) then

                spawn[#spawn + 1]     = candidate
                samples[#samples + 1] = candidate

                local hash = _getHash(candidate.x, candidate.y)

                if not hashTable[hash] then hashTable[hash] = {} end
                _insert(hashTable[hash], candidate)

                pointFound = true

                --Can be ommited
                if #samples % 2 == 1 then
                    love.graphics.setColor(1, 1, 0, .1)
                    for k = 0, 1920, hashSize do
                        love.graphics.line(k, 0, k, 1080)
                    end
                    for j = 0, 1080, hashSize do
                        love.graphics.line(0, j, 1920, j)
                    end

                    love.graphics.setColor(.3, .3, .3, .3)
                    for key, pos in ipairs(samples) do
                        local c = _sin((_floor(key * .01) / 20) % 100)

                        love.graphics.setColor(.1 + _sin(c), .1, 1 - _sin(c), .3)
                        love.graphics.circle("fill", pos.x, pos.y, pos.r)
                        love.graphics.setColor(.4, .4, .4, .3)
                        love.graphics.circle("line", pos.x, pos.y, pos.r)
                    end
                    love.graphics.setColor(1, 1, 1, 1)
                    love.graphics.print("Valid: " .. tostring(#samples) .. " / Remaining: " .. #spawn)
                    love.graphics.present()
                    love.graphics.clear(0, 0, 0, 0.1)
                end
                -- End ommited

                break
            end
        end

        if not pointFound then
            _remove(spawn, spawnIdx)
        end

    end
    --
    -- Error checking (brute force, slow)
    --
    -- for key1, value in pairs(samples) do
    --     for key2, point in pairs(samples) do
    --         if key1 ~= key2 then
    --             local r = _max(point.r, value.r)
    --             if _dist2(point, value) < r * r then
    --                 point.c = 1 -- signals color change in main.lua
    --                 print("Error", _dist2(point, value), r * r)
    --             end
    --         end
    --     end
    -- end
    return samples
end

---@class Poisson
local poisson = {
    ApplySettings = ApplySettings,
    GetSettings   = GetSettings,
    Generate      = Generate
}
return poisson
